'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.fireOnSequence = exports.dispatchActionWhen = exports.SEQ = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };
/**
 * Created by andrew on 10/13/16.
 */
/**
 * Schedules actions with { meta: { delay: N } } to be delayed by N milliseconds.
 * Makes `dispatch` return a function to cancel the timeout in this case.
 */


var _fluxStandardAction = require('flux-standard-action');

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var localNS = Symbol('SEQ');
var symNext = Symbol('next');
var symRestart = Symbol('restart');

function tagResultFunction(fn, tokenDescription, replacement) {
  fn.tag = localNS;
  if (replacement) {
    tokenDescription = fn.toString().replace(tokenDescription, replacement);
  }
  fn.toString = function () {
    return tokenDescription;
  };
  return fn;
}

function restart(tokens) {
  tokens.forEach(function (t) {
    return t(symRestart);
  });
}

var singleSequenceCreator = function singleSequenceCreator(token) {
  (0, _invariant2.default)(token, 'Token expected for SEQ.SINGLE');

  var tokenStr = null;

  switch (typeof token === 'undefined' ? 'undefined' : _typeof(token)) {
    case 'string':
      tokenStr = token;
      break;

    case 'function':
      if (token.tag === localNS) {
        return token;
      }
      tokenStr = token.toString();
      break;

    case 'object':
      {
        // TODO: make shallow or deep comparison
        if (typeof token.type == 'string') {
          tokenStr = token.type;
        }
      }
      break;
  }

  (0, _invariant2.default)(tokenStr, 'Token invalid type for SEQ.SINGLE');

  var result = function result(a) {
    return a !== symRestart && (a.type == tokenStr ? symNext : false);
  };
  var tokenDescription = 'SINGLE:(' + tokenStr + ')';
  return tagResultFunction(result, tokenDescription);
};

var timesSequenceCreator = function timesSequenceCreator(token, times) {
  var _ref = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  var strict = _ref.strict;

  (0, _invariant2.default)(token, 'Token expected for SEQ.TIMES');

  var normalizedToken = singleSequenceCreator(token);

  var tokenDescription = 'TIMES:(' + normalizedToken.toString() + ' x ' + times + ')';

  var count = 0;
  var result = function result(a) {
    if (a === symRestart) {
      count = 0;
      normalizedToken(symRestart);
      return false;
    }
    var result = normalizedToken(a);
    if (result === true) {
      return true;
    }

    if (result === symNext) {
      count = ++count % times;
      if (count === 0) {
        return symNext;
      } else {
        return true;
      }
    }

    if (strict) {
      count = 0;
      normalizedToken(symRestart);
      return false;
    }
    return true;
  };

  return tagResultFunction(result, tokenDescription);
};

var timesStrictSequenceCreator = function timesStrictSequenceCreator(token, times) {
  (0, _invariant2.default)(token, 'Token expected for SEQ.TIMES_STRICT');

  return tagResultFunction(timesSequenceCreator(token, times, { strict: true }), 'TIMES', 'TIMES_STRICT');
};

var allSequenceCreator = function allSequenceCreator(tokens) {
  var _ref2 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref2.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for SEQ.ALL');

  var normalizedTokens = tokens.map(singleSequenceCreator);
  var tokenDescription = 'ALL:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var results = normalizedTokens.map(function (t) {
    return false;
  });
  var length = normalizedTokens.length;

  var result = function result(a) {
    if (a === symRestart) {
      results = normalizedTokens.map(function (t) {
        return false;
      });
      restart(normalizedTokens);
      return false;
    }

    var nextResults = normalizedTokens.map(function (t, k) {
      var result = t(a);

      if (result === symNext) {
        return symNext;
      }
      return results[k];
    }, results);

    var nextResultsReadyCount = nextResults.filter(function (i) {
      return i === symNext;
    }).length;
    if (nextResultsReadyCount === length) {
      results = normalizedTokens.map(function (t) {
        return false;
      });
      restart(normalizedTokens);
      return symNext;
    } else if (strict && nextResultsReadyCount - results.length !== 1) {
      restart(normalizedTokens);
      return false;
    }

    results = nextResults;
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var anySequenceCreator = function anySequenceCreator(tokens) {
  var _ref3 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref3.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for SEQ.ANY');

  var normalizedTokens = tokens.map(singleSequenceCreator);
  var tokenDescription = 'ANY:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var result = function result(a) {
    if (a === symRestart) {
      restart(normalizedTokens);
      return false;
    }
    if (normalizedTokens.filter(function (t) {
      return t(a) === symNext;
    }).length) {
      restart(normalizedTokens);
      return symNext;
    }
    if (strict) {
      restart(normalizedTokens);
      return false;
    }
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var queueSequenceCreator = function queueSequenceCreator(tokens) {
  var _ref4 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref4.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for SEQ.QUEUE');

  var normalizedTokens = tokens.map(singleSequenceCreator);
  var tokenDescription = 'QUEUE:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var length = normalizedTokens.length;
  var count = 0;

  var result = function result(a) {
    if (a === symRestart) {
      count = 0;
      restart(normalizedTokens);
      return false;
    }
    var result = normalizedTokens[count](a);
    if (result === true) {
      return true;
    }

    if (result === symNext) {
      count = ++count % length;

      if (count === 0) {
        restart(normalizedTokens);
        return symNext;
      } else {
        return true;
      }
    }

    if (strict) {
      count = 0;
      restart(normalizedTokens);
      return false;
    }
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var queueStrictSequenceCreator = function queueStrictSequenceCreator(tokens) {
  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for SEQ.QUEUE_STRICT');

  return tagResultFunction(queueSequenceCreator(tokens, { strict: true }), 'QUEUE', 'QUEUE_STRICT');
};

var sequenceApi = {

  SINGLE: singleSequenceCreator,
  TIMES: timesSequenceCreator,
  TIMES_STRICT: timesStrictSequenceCreator,
  ALL: allSequenceCreator,
  ANY: anySequenceCreator,
  /**
   * @deprecated
   */
  SEQUENCE: queueSequenceCreator,
  QUEUE: queueSequenceCreator,
  /**
   * @deprecated
   */
  SEQUENCE_STRICT: queueStrictSequenceCreator,
  QUEUE_STRICT: queueStrictSequenceCreator
};

/**
 * @deprecated: use 'sequenceBuilderCb(s => ...) 'callback in 'dispatchWhen'
 * @type {{SINGLE: singleSequenceCreator, TIMES: timesSequenceCreator, TIMES_STRICT: timesStrictSequenceCreator, ALL: allSequenceCreator, ANY: anySequenceCreator, SEQUENCE: queueSequenceCreator, QUEUE: queueSequenceCreator, SEQUENCE_STRICT: queueStrictSequenceCreator, QUEUE_STRICT: queueStrictSequenceCreator}}
 */
var SEQ = exports.SEQ = sequenceApi;

var sequences = [];

var registerSequence = function registerSequence(fn) {

  sequences.push(fn);

  // unregister
  return function () {
    var idx = sequences.indexOf(fn);
    if (idx >= 0) {
      return sequences.splice(idx, 1);
    }
  };
};

var dispatchActionWhen = exports.dispatchActionWhen = function dispatchActionWhen(action, sequenceBuilderCb, _ref5) {
  var once = _ref5.once;

  return function (dispatch) {
    var sequence = sequenceBuilderCb(sequenceApi);
    (0, _invariant2.default)(sequence && typeof sequence === 'function', 'dispatchWhen expects sequenceBuilderCb to return compiled sequence (function)');

    var seqKey = 'SEQUENCE: ' + sequence + ' => ' + (action.type || action.toString());

    var unregister = registerSequence(function (a) {

      var result = sequence(a);
      if (result !== symNext) {
        return;
      }

      console.log(seqKey + ': RESOLVED'); // closure

      var unregisterFn = function unregisterFn() {
        unregister && unregister();
        unregister = null;
      };

      if (!once) {
        sequence(symRestart);
      }

      if (typeof action == 'function') {
        dispatch(action(unregisterFn));
      } else {
        dispatch(action);
      }

      if (once && unregister) {
        unregisterFn();
        unregisterFn = null;
      }
    });
  };
};

/**
 * @deprecated: use dispatchWhen
 * @param action
 * @param sequence
 * @param once
 * @returns {function(*)}
 */
var fireOnSequence = exports.fireOnSequence = function fireOnSequence(action, sequence, _ref6) {
  var once = _ref6.once;

  return dispatchActionWhen(action, function (SEQ) {
    return Array.isArray(sequence) ? SEQ.SEQUENCE(sequence) : SEQ.SINGLE(sequence);
  }, { once: once });
};

exports.default = function (store) {
  return function (next) {
    return function (action) {
      var result = next(action);

      if (sequences.length && (0, _fluxStandardAction.isFSA)(action)) {
        sequences.forEach(function (s) {
          return s(action);
        });
      }

      return result;
    };
  };
};
